@@ -22,6 +22,8 @@
 
 #include <ell/util.h>
 #include <ell/test.h>
+#include <ell/strv.h>
+#include <ell/utf8.h>
 #include <ell/queue.h>
 #include <ell/hashmap.h>
 #include <ell/string.h>
@@ -46,8 +48,11 @@
 #include <ell/key.h>
 #include <ell/pkcs5.h>
 #include <ell/file.h>
-
+#include <ell/dir.h>
+#include <ell/net.h>
 #include <ell/netlink.h>
 #include <ell/genl.h>
 #include <ell/dbus.h>
 #include <ell/dbus-service.h>
+#include <ell/dbus-client.h>
+#include <ell/dhcp.h>
